#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

module viscosity_mod
!
!  This module should be renamed "global_deriv_mod.F90"
!
!  It is a collection of derivative operators that must be applied to the field
!  over the sphere (as opposed to derivative operators that can be applied element
!  by element)
!
!
use thread_mod
use kinds, only : real_kind, iulog
use dimensions_mod, only : np, nlev,qsize,nelemd
use hybrid_mod, only : hybrid_t, config_thread_region
use parallel_mod, only : parallel_t
use element_mod, only : element_t
use derivative_mod, only : derivative_t, laplace_sphere_wk, vlaplace_sphere_wk, vorticity_sphere, derivinit, divergence_sphere
use edgetype_mod, only : EdgeBuffer_t
use edge_mod, only : edgevpack, edgerotate, edgevunpack, edgevunpackmin, &
    edgevunpackmax, initEdgeBuffer, FreeEdgeBuffer, edgeSunpackmax, edgeSunpackmin, edgeSpack
use bndry_mod, only : bndry_exchangev, bndry_exchangeS, bndry_exchangeS_start, bndry_exchangeS_finish
use control_mod, only : hypervis_scaling, nu, nu_div

implicit none
save

public :: biharmonic_wk
#ifdef _PRIM
public :: biharmonic_wk_scalar
public :: biharmonic_wk_scalar_minmax
public :: neighbor_minmax, neighbor_minmax_start, neighbor_minmax_finish
#endif

!
! compute vorticity/divergence and then project to make continious
! high-level routines uses only for I/O
public :: compute_zeta_C0
public :: compute_div_C0
interface compute_zeta_C0
    module procedure compute_zeta_C0_hybrid       ! hybrid version
    module procedure compute_zeta_C0_par          ! single threaded
end interface
interface compute_div_C0
    module procedure compute_div_C0_hybrid
    module procedure compute_div_C0_par
end interface

public :: compute_zeta_C0_contra    ! for older versions of sweq which carry
public :: compute_div_C0_contra     ! velocity around in contra-coordinates

public :: test_ibyp


type (EdgeBuffer_t)          :: edge1

contains

#ifdef _PRIM
subroutine biharmonic_wk(elem,pstens,ptens,vtens,deriv,edge3,hybrid,nt,nets,nete)
#else
subroutine biharmonic_wk(elem,ptens,vtens,deriv,edge3,hybrid,nt,nets,nete)
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute weak biharmonic operator
!    input:  h,v (stored in elem()%, in lat-lon coordinates
!    output: ptens,vtens  overwritten with weak biharmonic of h,v (output in lat-lon coordinates)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(inout), target :: elem(:)
integer :: nt,nets,nete
real (kind=real_kind), dimension(np,np,2,nlev,nets:nete)  :: vtens
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: ptens
type (EdgeBuffer_t)  , intent(inout) :: edge3
type (derivative_t)  , intent(in) :: deriv
#ifdef _PRIM
real (kind=real_kind), dimension(np,np,nets:nete) :: pstens
#endif

! local
integer :: k,kptr,i,j,ie,ic
real (kind=real_kind), dimension(:,:), pointer :: rspheremv
real (kind=real_kind), dimension(np,np) :: lap_ps
real (kind=real_kind), dimension(np,np,nlev) :: T
real (kind=real_kind), dimension(np,np,2) :: v
real (kind=real_kind) ::  nu_ratio1,nu_ratio2
logical var_coef1

   !if tensor hyperviscosity with tensor V is used, then biharmonic operator is (\grad\cdot V\grad) (\grad \cdot \grad)
   !so tensor is only used on second call to laplace_sphere_wk
   var_coef1 = .true.
   if(hypervis_scaling > 0)  var_coef1= .false.

   ! note: there is a scaling bug in the treatment of nu_div
   ! nu_ratio is applied twice, once in each laplace operator
   ! so in reality:   nu_div_actual = (nu_div/nu)**2 nu
   ! We should fix this, but it requires adjusting all CAM defaults
   nu_ratio1=1
   nu_ratio2=1
   if (nu_div/=nu) then
      if(hypervis_scaling /= 0) then
         ! we have a problem with the tensor in that we cant seperate
         ! div and curl components.  So we do, with tensor V:
         ! nu * (del V del ) * ( nu_ratio * grad(div) - curl(curl))
         nu_ratio1=(nu_div/nu)**2   ! preserve buggy scaling
         nu_ratio2=1
      else
         nu_ratio1=nu_div/nu
         nu_ratio2=nu_div/nu
      endif
   endif


   do ie=nets,nete

#ifdef _PRIM
      ! should filter lnps + PHI_s/RT?
      !pstens(:,:,ie)=laplace_sphere_wk(elem(ie)%state%ps_v(:,:,nt),deriv,elem(ie),var_coef=var_coef1)
      call laplace_sphere_wk(elem(ie)%state%ps_v(:,:,nt),deriv,elem(ie),pstens(:,:,ie),var_coef=var_coef1)
#endif

#if (defined COLUMN_OPENMP)
!$omp parallel do private(k, j, i)
#endif
      do k=1,nlev
         do j=1,np
            do i=1,np
#ifdef _PRIM
               T(i,j,k)=elem(ie)%state%T(i,j,k,nt)
#elif defined _PRIMDG
               T(i,j,k)=elem(ie)%state%p(i,j,k,nt) + elem(ie)%state%phis(i,j)
#else
               ! filter surface height, not thickness
               T(i,j,k)=elem(ie)%state%p(i,j,k,nt) + elem(ie)%state%ps(i,j)
#endif
            enddo
         enddo

         !ptens(:,:,k,ie)=laplace_sphere_wk(T(:,:,k),deriv,elem(ie),var_coef=var_coef1)
         call laplace_sphere_wk(T(:,:,k),deriv,elem(ie),ptens(:,:,k,ie),var_coef=var_coef1)
         !vtens(:,:,:,k,ie)=vlaplace_sphere_wk(elem(ie)%state%v(:,:,:,k,nt),deriv,&
         !     elem(ie),var_coef=var_coef1,nu_ratio=nu_ratio1)
         call vlaplace_sphere_wk(elem(ie)%state%v(:,:,:,k,nt),deriv,&
              elem(ie),vtens(:,:,:,k,ie),var_coef=var_coef1,nu_ratio=nu_ratio1)

      enddo
      kptr=0
      call edgeVpack(edge3, ptens(:,:,:,ie),nlev,kptr,ie)
      kptr=nlev
      call edgeVpack(edge3, vtens(:,:,1,:,ie),nlev,kptr,ie)
      kptr=2*nlev
      call edgeVpack(edge3, vtens(:,:,2,:,ie),nlev,kptr,ie)

#ifdef _PRIM
      kptr=3*nlev
      call edgeVpack(edge3, pstens(:,:,ie),1,kptr,ie)
#endif
   enddo

   call bndry_exchangeV(hybrid,edge3)

   do ie=nets,nete
      rspheremv     => elem(ie)%rspheremp(:,:)

      kptr=0
      call edgeVunpack(edge3, ptens(:,:,:,ie), nlev, kptr, ie)
      kptr=nlev
      call edgeVunpack(edge3, vtens(:,:,1,:,ie), nlev, kptr, ie)
      kptr=2*nlev
      call edgeVunpack(edge3, vtens(:,:,1,:,ie), nlev, kptr, ie)

      ! apply inverse mass matrix, then apply laplace again
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k, j, i, v)
#endif
      do k=1,nlev
         do j=1,np
            do i=1,np
               T(i,j,k)=rspheremv(i,j)*ptens(i,j,k,ie)
               v(i,j,1)=rspheremv(i,j)*vtens(i,j,1,k,ie)
               v(i,j,2)=rspheremv(i,j)*vtens(i,j,2,k,ie)
            enddo
         enddo
         !ptens(:,:,k,ie)=laplace_sphere_wk(T(:,:,k),deriv,elem(ie),var_coef=.true.)
         call laplace_sphere_wk(T(:,:,k),deriv,elem(ie),ptens(:,:,k,ie),var_coef=.true.)
         !vtens(:,:,:,k,ie)=vlaplace_sphere_wk(v(:,:,:),deriv,elem(ie),var_coef=.true.,&
         !     nu_ratio=nu_ratio2)
         call vlaplace_sphere_wk(v(:,:,:),deriv,elem(ie),vtens(:,:,:,k,ie),var_coef=.true.,&
              nu_ratio=nu_ratio2)
      enddo

#ifdef _PRIM
      kptr=3*nlev
      call edgeVunpack(edge3, pstens(:,:,ie), 1, kptr,ie)
      ! apply inverse mass matrix, then apply laplace again
      lap_ps(:,:)=rspheremv(:,:)*pstens(:,:,ie)
      !pstens(:,:,ie)=laplace_sphere_wk(lap_ps,deriv,elem(ie),var_coef=.true.)
      call laplace_sphere_wk(lap_ps,deriv,elem(ie),pstens(:,:,ie),var_coef=.true.)
#endif

   enddo
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
!$OMP BARRIER
#endif
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end subroutine

#ifdef _PRIM
#define SW_BIHARMONIC_WK_DP3D_SLAVE
#define SW_BIHARMONIC_WK_DP3D_PACK_UNPACK
subroutine biharmonic_wk_dp3d(elem,dptens,ptens,vtens,deriv,edge3,hybrid,nt,nets,nete)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute weak biharmonic operator
!    input:  h,v (stored in elem()%, in lat-lon coordinates
!    output: ptens,vtens  overwritten with weak biharmonic of h,v (output in lat-lon coordinates)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifdef SW_BIHARMONIC_WK_DP3D_SLAVE
	use physical_constants, only :rrearth
  use control_mod, only : hypervis_scaling, hypervis_power
#endif
#ifdef SW_BIHARMONIC_WK_DP3D_PACK_UNPACK
  use control_mod, only : nu, nu_div, nu_s, hypervis_order, hypervis_subcycle, nu_p, nu_top, psurf_vis, north, south, east, west, swest
	use spmd_utils, only: iam
  use kinds, only : real_kind, int_kind, iulog
  use dimensions_mod, only : np, np, nlev, max_corner_elem
#endif
use perf_mod, only: t_startf, t_stopf, t_barrierf, t_adj_detailf ! _EXTERNAL

type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(inout), target :: elem(:)
integer :: nt,nets,nete
real (kind=real_kind), dimension(np,np,2,nlev,nets:nete)  :: vtens
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: ptens,dptens
type (EdgeBuffer_t)  , intent(inout) :: edge3
type (derivative_t)  , intent(in) :: deriv

! local
integer :: k,kptr,ie
real (kind=real_kind), dimension(:,:), pointer :: rspheremv
real (kind=real_kind), dimension(np,np) :: tmp
real (kind=real_kind), dimension(np,np,2) :: v
real (kind=real_kind) :: nu_ratio1, nu_ratio2
logical var_coef1
#ifdef SW_BIHARMONIC_WK_DP3D_SLAVE
	!SUNWAY
	!struct declearation
	type str_g
		integer(kind=8)	:: ptr_e_T,ptr_e_dp3d,ptr_ptens,ptr_dptens,& !lap
		ptr_e_variable_hyperviscosity,ptr_e_tensorVisc,&
		ptr_e_Dinv,ptr_d_Dvv,ptr_e_spheremp,&
		ptr_e_v,ptr_vtens,ptr_e_vec_cart,ptr_e_metdet,	&!vlap
		ptr_e_rmetdet,ptr_e_Dvv,ptr_e_D,ptr_e_mp,ptr_e_metinv
		real(kind=8)	:: hypervis_power,hypervis_scaling,rrearth,& !lap
		nu_ratio !vlap
		integer var_coef
		integer :: nets,nete,nlev,np,len_elem,len_ptens,len_dptens,len_vtens
	end type str_g
	type str_h
		integer(kind=8)	:: ptr_e_rspheremp,ptr_ptens,ptr_dptens,&!lap
			ptr_e_variable_hyperviscosity,ptr_e_tensorVisc,&
			ptr_e_Dinv,ptr_d_Dvv,ptr_e_spheremp,&
			ptr_vtens,ptr_e_vec_cart,ptr_e_metdet,	&!vlap
			ptr_e_rmetdet,ptr_e_Dvv,ptr_e_D,ptr_e_mp,ptr_e_metinv
		real(kind=8)	:: hypervis_power,hypervis_scaling,rrearth,&!lap
			nu_ratio2 !vlap
		integer var_coef
		integer :: nets,nete,nlev,np,len_elem,len_ptens,len_dptens,len_vtens;
	end type str_h
	!slave function declearation
	external	:: slave_advance_hypervis_dp_kernelg
	external	:: slave_advance_hypervis_dp_kernelh
	! inputs for slave function
	type(str_g)	:: strpar_g
	type(str_h)	:: strpar_h
#endif
  !-------------------for edgeVpack slave kernel---------------------------
#ifdef SW_BIHARMONIC_WK_DP3D_PACK_UNPACK
  type param_t
    integer(kind=8) :: buf
    integer(kind=8) :: receive
    integer(kind=8) :: addrs(3)
    integer(kind=8) :: putmap_addr
    integer(kind=8) :: getmap_addr
    integer(kind=8) :: reverse_addr
    integer(kind=8) :: iwesn_addr
    integer(kind=4) :: np, nlev, nets, nete, max_corner_elem, iam
  end type
  type(param_t) :: param
  integer (kind=int_kind), dimension(:,:), allocatable :: putmap
  integer (kind=int_kind), dimension(:,:), allocatable :: getmap
  integer (kind=int_kind), dimension(:,:), allocatable :: iwesn
  integer (kind=int_kind), dimension(:,:), allocatable :: reverse
  external :: slave_edgevpack3_parallel
  external :: slave_edgevunpack3_parallel
	!local
	integer i
#endif

   !if tensor hyperviscosity with tensor V is used, then biharmonic operator is (\grad\cdot V\grad) (\grad \cdot \grad)
   !so tensor is only used on second call to laplace_sphere_wk
   var_coef1 = .true.
   if(hypervis_scaling > 0)    var_coef1 = .false.

   ! note: there is a scaling bug in the treatment of nu_div
   ! nu_ratio is applied twice, once in each laplace operator
   ! so in reality:   nu_div_actual = (nu_div/nu)**2 nu
   ! We should fix this, but it requires adjusting all CAM defaults
   nu_ratio1=1
   nu_ratio2=1
   if (nu_div/=nu) then
      if(hypervis_scaling /= 0) then
         ! we have a problem with the tensor in that we cant seperate
         ! div and curl components.  So we do, with tensor V:
         ! nu * (del V del ) * ( nu_ratio * grad(div) - curl(curl))
         nu_ratio1=(nu_div/nu)**2   ! preserve buggy scaling
         nu_ratio2=1
      else
         nu_ratio1=nu_div/nu
         nu_ratio2=nu_div/nu
      endif
   endif

		call t_startf("kernel-g")
#ifdef SW_BIHARMONIC_WK_DP3D_SLAVE
		!SUNWAY KernelG
		!lap
		strpar_g%ptr_e_T=loc(elem(1)%state%T(1,1,1,nt))			!str_par%ptr_T
		strpar_g%ptr_e_dp3d=loc(elem(1)%state%dp3d(1,1,1,nt))
		strpar_g%ptr_ptens=loc(ptens(1,1,1,1))
		strpar_g%ptr_dptens=loc(dptens(1,1,1,1))
		strpar_g%ptr_e_variable_hyperviscosity=loc(elem(1)%variable_hyperviscosity(1,1))
		strpar_g%ptr_e_tensorVisc=loc(elem(1)%tensorVisc(1,1,1,1))
		strpar_g%ptr_e_Dinv=loc(elem(1)%Dinv(1,1,1,1))
		strpar_g%ptr_d_Dvv=loc(deriv%Dvv(1,1))
		strpar_g%ptr_e_spheremp=loc(elem(1)%spheremp(1,1))
		strpar_g%hypervis_power=hypervis_power
		strpar_g%hypervis_scaling=hypervis_scaling
		strpar_g%rrearth=rrearth
		if(var_coef1) then
			strpar_g%var_coef=1
		else
			strpar_g%var_coef=0
		endif
		strpar_g%nets=nets
		strpar_g%nete=nete
		strpar_g%nlev=nlev
		strpar_g%np=np
		strpar_g%len_elem=(loc(elem(2))-loc(elem(1)))/8
		strpar_g%len_ptens=(loc(ptens(1,1,1,2))-loc(ptens(1,1,1,1)))/8
		strpar_g%len_dptens=(loc(dptens(1,1,1,2))-loc(dptens(1,1,1,1)))/8
		!vlap-0
		strpar_g%ptr_e_vec_cart=loc(elem(1)%vec_sphere2cart(1,1,1,1))
		strpar_g%ptr_e_v=loc(elem(1)%state%v(1,1,1,1,nt))
		strpar_g%ptr_vtens=loc(vtens(1,1,1,1,1))
		!vlap
		strpar_g%ptr_e_metdet=loc(elem(1)%metdet(1,1))
		strpar_g%ptr_e_rmetdet=loc(elem(1)%rmetdet(1,1))
		strpar_g%ptr_e_D=loc(elem(1)%D(1,1,1,1))
		strpar_g%ptr_e_mp=loc(elem(1)%mp(1,1))
		strpar_g%ptr_e_metinv=loc(elem(1)%metinv(1,1,1,1))
		strpar_g%nu_ratio=nu_ratio1
		strpar_g%len_vtens=(loc(vtens(1,1,1,1,2))-loc(vtens(1,1,1,1,1)))/8
		call athread_spawn(slave_advance_hypervis_dp_kernelg,strpar_g)
		call athread_join()

#else
   do ie=nets,nete

#if (defined COLUMN_OPENMP)
!$omp parallel do default(shared), private(k,tmp)
#endif
      do k=1,nlev
         tmp=elem(ie)%state%T(:,:,k,nt)
         !ptens(:,:,k,ie)=laplace_sphere_wk(tmp,deriv,elem(ie),var_coef=var_coef1)
         call laplace_sphere_wk(tmp,deriv,elem(ie),ptens(:,:,k,ie),var_coef=var_coef1)
         tmp=elem(ie)%state%dp3d(:,:,k,nt)
         !dptens(:,:,k,ie)=laplace_sphere_wk(tmp,deriv,elem(ie),var_coef=var_coef1)
         call laplace_sphere_wk(tmp,deriv,elem(ie),dptens(:,:,k,ie),var_coef=var_coef1)
         !vtens(:,:,:,k,ie)=vlaplace_sphere_wk(elem(ie)%state%v(:,:,:,k,nt),deriv,elem(ie),&
         !     var_coef=var_coef1,nu_ratio=nu_ratio1)
         call vlaplace_sphere_wk(elem(ie)%state%v(:,:,:,k,nt),deriv,elem(ie),&
              vtens(:,:,:,k,ie), var_coef=var_coef1,nu_ratio=nu_ratio1)
      enddo
   enddo
#endif
		call t_stopf("kernel-g")

#ifdef SW_BIHARMONIC_WK_DP3D_PACK_UNPACK
		call t_startf("kernel-g-pack")
!		do ie=nets,nete
!			kptr=0
!			call edgeVpack(edge3, ptens(1,1,1,ie),nlev,kptr,ie)
!			kptr=nlev
!			call edgeVpack(edge3, vtens(1,1,1,1,ie),2*nlev,kptr,ie)
!			kptr=3*nlev
!			call edgeVpack(edge3, dptens(1,1,1,ie),nlev,kptr,ie)
!		enddo
	  allocate(putmap(swest:swest+4*max_corner_elem-1,nets:nete))
	  allocate(getmap(swest:swest+4*max_corner_elem-1,nets:nete))
	  allocate(iwesn(1:4,nets:nete))
	  allocate(reverse(1:4,nets:nete))
        do ie = nets, nete
          iwesn(1, ie)=edge3%putmap(west, ie)
          iwesn(2, ie)=edge3%putmap(east, ie)
          iwesn(3, ie)=edge3%putmap(south, ie)
          iwesn(4, ie)=edge3%putmap(north, ie)
         do i = swest, swest+4*max_corner_elem-1
           putmap(i, ie)=edge3%putmap(i, ie)
	   enddo
          if ( edge3%reverse(west, ie) ) then
            reverse(1, ie)=1
          else
            reverse(1, ie)=0
          end if
          if ( edge3%reverse(east, ie) ) then
            reverse(2, ie)=1
          else
            reverse(2, ie)=0
          end if
          if ( edge3%reverse(south, ie) ) then
            reverse(3, ie)=1
          else
            reverse(3, ie)=0
          end if
          if ( edge3%reverse(north, ie) ) then
            reverse(4, ie)=1
          else
            reverse(4, ie)=0
          end if
        enddo
        param%addrs(1)=loc(ptens)
        param%addrs(2)=loc(vtens)
        param%addrs(3)=loc(dptens)
!        param%buf = loc(retbuf)
        param%buf = loc(edge3%buf)
        param%receive = loc(edge3%receive)
	  param%getmap_addr = loc(getmap)
	  param%iwesn_addr =loc(iwesn)
	  param%putmap_addr = loc(putmap)
	  param%reverse_addr =loc(reverse)
        param%np = np
        param%nlev = nlev
        param%nets = nets
        param%nete = nete
        param%max_corner_elem = max_corner_elem
        param%iam = iam

        call athread_spawn(slave_edgevpack3_parallel, param)
        call athread_join()
		call t_stopf("kernel-g-pack")

		call t_startf("bndry_exchangeV-2")
		call bndry_exchangeV(hybrid,edge3)
		call t_stopf("bndry_exchangeV-2")

		call t_startf("kernel-h-unpack")
!		do ie=nets,nete
!			kptr=0
!			call edgeVunpack(edge3, ptens(1,1,1,ie), nlev, kptr, ie)
!			kptr=nlev
!			call edgeVunpack(edge3, vtens(1,1,1,1,ie), 2*nlev, kptr, ie)
!			kptr=3*nlev
!			call edgeVunpack(edge3, dptens(1,1,1,ie), nlev, kptr, ie)
!		enddo
	  do ie = nets, nete
	    do i = swest, swest+4*max_corner_elem-1
            getmap(i, ie) = edge3%getmap(i, ie)
	    enddo
	    iwesn(1, ie) = edge3%getmap(west,ie)
          iwesn(2, ie) = edge3%getmap(east,ie)
	    iwesn(3, ie) = edge3%getmap(south,ie)
	    iwesn(4, ie) = edge3%getmap(north,ie)
	  enddo

       call athread_spawn(slave_edgevunpack3_parallel, param)
       call athread_join()

	  deallocate(getmap)
	  deallocate(putmap)
	  deallocate(iwesn)
	  deallocate(reverse)

		call t_stopf("kernel-h-unpack")
#else
		call t_startf("kernel-g-pack")
		do ie=nets,nete
			kptr=0
			call edgeVpack(edge3, ptens(1,1,1,ie),nlev,kptr,ie)
			kptr=nlev
			call edgeVpack(edge3, vtens(1,1,1,1,ie),2*nlev,kptr,ie)
			kptr=3*nlev
			call edgeVpack(edge3, dptens(1,1,1,ie),nlev,kptr,ie)
		enddo
		call t_stopf("kernel-g-pack")

		call t_startf("bndry_exchangeV-2")
		call bndry_exchangeV(hybrid,edge3)
		call t_stopf("bndry_exchangeV-2")

		call t_startf("kernel-h-unpack")
		do ie=nets,nete
			kptr=0
			call edgeVunpack(edge3, ptens(1,1,1,ie), nlev, kptr, ie)
			kptr=nlev
			call edgeVunpack(edge3, vtens(1,1,1,1,ie), 2*nlev, kptr, ie)
			kptr=3*nlev
			call edgeVunpack(edge3, dptens(1,1,1,ie), nlev, kptr, ie)
		enddo
		call t_stopf("kernel-h-unpack")
#endif

		call t_startf("kernel-h")
#ifdef SW_BIHARMONIC_WK_DP3D_SLAVE
		strpar_h%ptr_e_rspheremp=loc(elem(1)%rspheremp(1,1))
		strpar_h%ptr_ptens=loc(ptens(1,1,1,1))
		strpar_h%ptr_dptens=loc(dptens(1,1,1,1))
		strpar_h%ptr_e_variable_hyperviscosity=loc(elem(1)%variable_hyperviscosity(1,1))
		strpar_h%ptr_e_tensorVisc=loc(elem(1)%tensorVisc(1,1,1,1))
		strpar_h%ptr_e_Dinv=loc(elem(1)%Dinv(1,1,1,1))
		strpar_h%ptr_d_Dvv=loc(deriv%Dvv(1,1))
		strpar_h%ptr_e_spheremp=loc(elem(1)%spheremp(1,1))
		strpar_h%hypervis_power=hypervis_power
		strpar_h%hypervis_scaling=hypervis_scaling
		strpar_h%rrearth=rrearth
		strpar_h%var_coef=1 !true
		strpar_h%nets=nets
		strpar_h%nete=nete
		strpar_h%nlev=nlev
		strpar_h%np=np
		strpar_h%len_elem=(loc(elem(2))-loc(elem(1)))/8
		strpar_h%len_ptens=(loc(ptens(1,1,1,2))-loc(ptens(1,1,1,1)))/8
		strpar_h%len_dptens=(loc(dptens(1,1,1,2))-loc(dptens(1,1,1,1)))/8
		!vlap-0
		strpar_h%ptr_e_vec_cart=loc(elem(1)%vec_sphere2cart(1,1,1,1))
		strpar_h%ptr_vtens=loc(vtens(1,1,1,1,1))
		!vlap
		strpar_h%ptr_e_metdet=loc(elem(1)%metdet(1,1))
		strpar_h%ptr_e_rmetdet=loc(elem(1)%rmetdet(1,1))
		strpar_h%ptr_e_D=loc(elem(1)%D(1,1,1,1))
		strpar_h%ptr_e_mp=loc(elem(1)%mp(1,1))
		strpar_h%ptr_e_metinv=loc(elem(1)%metinv(1,1,1,1))
		strpar_h%nu_ratio2=nu_ratio2
		strpar_h%len_vtens=(loc(vtens(1,1,1,1,2))-loc(vtens(1,1,1,1,1)))/8
		call athread_spawn(slave_advance_hypervis_dp_kernelh,strpar_h)
		call athread_join()

#else
   do ie=nets,nete
      rspheremv     => elem(ie)%rspheremp(:,:)

      ! apply inverse mass matrix, then apply laplace again
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k,v,tmp)
#endif
      do k=1,nlev
         tmp(:,:)=rspheremv(:,:)*ptens(:,:,k,ie)
         !ptens(:,:,k,ie)=laplace_sphere_wk(tmp,deriv,elem(ie),var_coef=.true.)
         call laplace_sphere_wk(tmp,deriv,elem(ie),ptens(:,:,k,ie),var_coef=.true.)
         tmp(:,:)=rspheremv(:,:)*dptens(:,:,k,ie)
         !dptens(:,:,k,ie)=laplace_sphere_wk(tmp,deriv,elem(ie),var_coef=.true.)
         call laplace_sphere_wk(tmp,deriv,elem(ie),dptens(:,:,k,ie),var_coef=.true.)

         v(:,:,1)=rspheremv(:,:)*vtens(:,:,1,k,ie)
         v(:,:,2)=rspheremv(:,:)*vtens(:,:,2,k,ie)
         !vtens(:,:,:,k,ie)=vlaplace_sphere_wk(v(:,:,:),deriv,elem(ie),&
         !     var_coef=.true.,nu_ratio=nu_ratio2)
         call vlaplace_sphere_wk(v(:,:,:),deriv,elem(ie),&
              vtens(:,:,:,k,ie),var_coef=.true.,nu_ratio=nu_ratio2)

      enddo
   enddo
#endif
		call t_stopf("kernel-h")
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
!$OMP BARRIER
#endif
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end subroutine

#define BIHARMONIC_SW
#ifdef BIHARMONIC_SW
  subroutine biharmonic_wk_scalar(elem,qtens,deriv,edgeq,hybrid,nets,nete)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! compute weak biharmonic operator
  !    input:  qtens = Q
  !    output: qtens = weak biharmonic of Q
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
    use physical_constants, only : rrearth
    use control_mod, only : hypervis_scaling, hypervis_power
    use control_mod, only : north, south, east, west, neast, nwest, seast, swest
    type (hybrid_t)      , intent(in) :: hybrid
    type (element_t)     , intent(inout), target :: elem(:)
    integer :: nets,nete
    real (kind=real_kind), dimension(np,np,nlev,qsize,nets:nete) :: qtens
    type (EdgeBuffer_t)  , intent(inout) :: edgeq
    type (derivative_t)  , intent(in) :: deriv

    ! local
    integer :: k,kptr,i,j,ie,ic,q,l,m,n
    real (kind=real_kind), dimension(np,np) :: lap_p
    logical var_coef1

    logical :: var_coef

    ! Local
    real(kind=real_kind) :: grads(np,np,2), oldgrads(np,np,2)

    !call gradient_sphere(s,deriv,elem%Dinv, grads)

    real(kind=real_kind) ::  dsdx00
    real(kind=real_kind) ::  dsdy00
    real(kind=real_kind) ::  v1(np,np),v2(np,np)

    real(kind=real_kind) ::  vtemp(np,np,2)
    integer :: reverse(4, nets:nete)

    integer :: step_elem
    external :: slave_biharmonic_wk_scalar
    external :: slave_biharmonic_wk_scalar_2
    type param_t
      integer*8 :: Dinv, rspheremp, spheremp, variable_hyperviscosity, tensorVisc, Qtens     \
          , Dvv, buf
      integer*8 :: putmap, reverse
      real(kind=real_kind) :: rrearth, hypervis_scaling, hypervis_power
      integer :: nets, nete, qsize, step_elem
    end type param_t
    type(param_t) :: param_s
    type param_2_t
      integer*8 :: Dinv, rspheremp, spheremp, variable_hyperviscosity, tensorVisc, Qtens     \
          , Dvv, receive
      integer*8 :: getmap
      real(kind=real_kind) :: rrearth, hypervis_scaling, hypervis_power
      integer :: nets, nete, qsize, step_elem
    end type param_2_t
    type(param_2_t) :: param_s2

    var_coef1 = .true.
    if(hypervis_scaling > 0)    var_coef1 = .false.

    call t_startf('sw_biharmonic_1_vis')
    do ie = nets, nete
      if (edgeq%reverse(west, ie)) then
        reverse(1, ie) = 1
      else
        reverse(1, ie) = 0
      endif
      if (edgeq%reverse(east, ie)) then
        reverse(2, ie) = 1
      else
        reverse(2, ie) = 0
      endif
      if (edgeq%reverse(south, ie)) then
        reverse(3, ie) = 1
      else
        reverse(3, ie) = 0
      endif
      if (edgeq%reverse(north, ie)) then
        reverse(4, ie) = 1
      else
        reverse(4, ie) = 0
      endif
    enddo

    param_s%Dinv = loc(elem(nets)%Dinv)
    param_s%rspheremp = loc(elem(nets)%rspheremp)
    param_s%spheremp = loc(elem(nets)%spheremp)
    param_s%variable_hyperviscosity = loc(elem(nets)%variable_hyperviscosity)
    param_s%tensorVisc = loc(elem(nets)%tensorVisc)
    param_s%Qtens = loc(Qtens)
    param_s%Dvv = loc(deriv%Dvv)
    param_s%buf = loc(edgeq%buf)
    param_s%putmap = loc(edgeq%putmap(1,nets))
    param_s%reverse = loc(reverse)
    param_s%rrearth = rrearth
    param_s%hypervis_scaling = hypervis_scaling
    param_s%hypervis_power = hypervis_power
    param_s%nets = nets
    param_s%nete = nete
    param_s%qsize = qsize
    param_s%step_elem = (loc(elem(nets+1)%Dinv) - loc(elem(nets)%Dinv))/8
    call athread_spawn(slave_biharmonic_wk_scalar, param_s)
    call athread_join()
    call t_stopf('sw_biharmonic_1_vis')

    call t_startf('sw_biharmonic_bndry_exchangeV_vis')
    call bndry_exchangeV(hybrid,edgeq)
    call t_stopf('sw_biharmonic_bndry_exchangeV_vis')

    call t_startf('sw_biharmonic_2_vis')
    param_s2%Dinv = loc(elem(nets)%Dinv)
    param_s2%rspheremp = loc(elem(nets)%rspheremp)
    param_s2%spheremp = loc(elem(nets)%spheremp)
    param_s2%variable_hyperviscosity = loc(elem(nets)%variable_hyperviscosity)
    param_s2%tensorVisc = loc(elem(nets)%tensorVisc)
    param_s2%Qtens = loc(Qtens)
    param_s2%Dvv = loc(deriv%Dvv)
    param_s2%receive = loc(edgeq%receive)
    param_s2%getmap = loc(edgeq%getmap(1,nets))
    param_s2%rrearth = rrearth
    param_s2%hypervis_scaling = hypervis_scaling
    param_s2%hypervis_power = hypervis_power
    param_s2%nets = nets
    param_s2%nete = nete
    param_s2%qsize = qsize
    param_s2%step_elem = (loc(elem(nets+1)%Dinv) - loc(elem(nets)%Dinv))/8
    call athread_spawn(slave_biharmonic_wk_scalar_2, param_s2)
    call athread_join()
    call t_stopf('sw_biharmonic_2_vis')
  end subroutine
#else
  subroutine biharmonic_wk_scalar(elem,qtens,deriv,edgeq,hybrid,nets,nete)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! compute weak biharmonic operator
  !    input:  qtens = Q
  !    output: qtens = weak biharmonic of Q
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type (hybrid_t)      , intent(in) :: hybrid
  type (element_t)     , intent(inout), target :: elem(:)
  integer :: nets,nete
  real (kind=real_kind), dimension(np,np,nlev,qsize,nets:nete) :: qtens
  type (EdgeBuffer_t)  , intent(inout) :: edgeq
  type (derivative_t)  , intent(in) :: deriv

  ! local
  integer :: k,kptr,i,j,ie,ic,q
  real (kind=real_kind), dimension(np,np) :: lap_p
  logical var_coef1

     !if tensor hyperviscosity with tensor V is used, then biharmonic operator is (\grad\cdot V\grad) (\grad \cdot \grad)
     !so tensor is only used on second call to laplace_sphere_wk
     var_coef1 = .true.
     if(hypervis_scaling > 0)    var_coef1 = .false.



     do ie=nets,nete
        do q=1,qsize
           do k=1,nlev    !  Potential loop inversion (AAM)
              lap_p(:,:)=qtens(:,:,k,q,ie)
  ! Original use of qtens on left and right hand sides caused OpenMP errors (AAM)
             !qtens(:,:,k,q,ie)=laplace_sphere_wk(lap_p,deriv,elem(ie),var_coef=var_coef1)
             call laplace_sphere_wk(lap_p,deriv,elem(ie),qtens(:,:,k,q,ie),var_coef=var_coef1)
           enddo
           kptr = nlev*(q-1)
           call edgeVpack(edgeq, qtens(:,:,1:nlev,q,ie),nlev,kptr,ie)
        enddo
     enddo

     call bndry_exchangeV(hybrid,edgeq)

     do ie=nets,nete
        ! apply inverse mass matrix, then apply laplace again
        do q=1,qsize
          kptr = nlev*(q-1)
          call edgeVunpack(edgeq, qtens(:,:,1:nlev,q,ie),nlev,kptr,ie)
          do k=1,nlev    !  Potential loop inversion (AAM)
             lap_p(:,:)=elem(ie)%rspheremp(:,:)*qtens(:,:,k,q,ie)
             ! qtens(:,:,k,q,ie)=laplace_sphere_wk(lap_p,deriv,elem(ie),var_coef=.true.)
             call laplace_sphere_wk(lap_p,deriv,elem(ie),qtens(:,:,k,q,ie),var_coef=.true.)
          enddo
        enddo
     enddo
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
  !$OMP BARRIER
#endif
#endif
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  end subroutine
#endif

subroutine biharmonic_wk_scalar_minmax(elem,qtens,deriv,edgeq,edgeminmax,hybrid,nets,nete,emin,emax)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute weak biharmonic operator
!    input:  qtens = Q
!    output: qtens = weak biharmonic of Q and Q element min/max
!
!    note: emin/emax must be initialized with Q element min/max.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(inout), target :: elem(:)
integer :: nets,nete
real (kind=real_kind), dimension(np,np,nlev,qsize,nets:nete) :: qtens
type (EdgeBuffer_t)  , intent(inout) :: edgeq
type (EdgeBuffer_t)  , intent(inout) :: edgeminmax
type (derivative_t)  , intent(in) :: deriv
real (kind=real_kind), intent(out), dimension(nlev,qsize,nets:nete) :: emin,emax

! local
integer :: k,kptr,i,j,ie,ic,q
real (kind=real_kind), dimension(np,np) :: lap_p
real (kind=real_kind) :: Qmin(np,np,nlev,qsize)
real (kind=real_kind) :: Qmax(np,np,nlev,qsize)
logical var_coef1

   !if tensor hyperviscosity with tensor V is used, then biharmonic operator is (\grad\cdot V\grad) (\grad \cdot \grad)
   !so tensor is only used on second call to laplace_sphere_wk
   var_coef1 = .true.
   if(hypervis_scaling > 0)    var_coef1 = .false.



   do ie=nets,nete
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k, q, lap_p)
#endif
      do q=1,qsize
      do k=1,nlev    !  Potential loop inversion (AAM)
         Qmin(:,:,k,q)=emin(k,q,ie)  ! need to set all values in element for
         Qmax(:,:,k,q)=emax(k,q,ie)  ! edgeVpack routine below
         lap_p(:,:) = qtens(:,:,k,q,ie)
! Original use of qtens on left and right hand sides caused OpenMP errors (AAM)
         ! qtens(:,:,k,q,ie)=laplace_sphere_wk(lap_p,deriv,elem(ie),var_coef=var_coef1)
         call laplace_sphere_wk(lap_p,deriv,elem(ie),qtens(:,:,k,q,ie),var_coef=var_coef1)
      enddo
      enddo
      call edgeVpack(edgeq, qtens(:,:,:,:,ie),qsize*nlev,0,ie)
      call edgeVpack(edgeq,Qmin,nlev*qsize,nlev*qsize,ie)
      call edgeVpack(edgeq,Qmax,nlev*qsize,2*nlev*qsize,ie)
   enddo

   call bndry_exchangeV(hybrid,edgeq)

   do ie=nets,nete
      do q=1,qsize
      do k=1,nlev
         Qmin(:,:,k,q)=emin(k,q,ie)  ! restore element data.  we could avoid
         Qmax(:,:,k,q)=emax(k,q,ie)  ! this by adding a "ie" index to Qmin/max
      enddo
      enddo
! WARNING - edgeVunpackMin/Max take second argument as input/ouput
      call edgeVunpack(edgeq, qtens(:,:,:,:,ie),qsize*nlev,0,ie)
      call edgeVunpackMin(edgeq, Qmin,qsize*nlev,qsize*nlev,ie)
      call edgeVunpackMax(edgeq, Qmax,qsize*nlev,2*qsize*nlev,ie)

      ! apply inverse mass matrix, then apply laplace again
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k, q, lap_p)
#endif
      do q=1,qsize
        do k=1,nlev    !  Potential loop inversion (AAM)
           lap_p(:,:)=elem(ie)%rspheremp(:,:)*qtens(:,:,k,q,ie)
           ! qtens(:,:,k,q,ie)=laplace_sphere_wk(lap_p,deriv,elem(ie),var_coef=.true.)
           call laplace_sphere_wk(lap_p,deriv,elem(ie),qtens(:,:,k,q,ie),var_coef=.true.)
           ! note: only need to consider the corners, since the data we packed was
           ! constant within each element
           emin(k,q,ie)=min(qmin(1,1,k,q),qmin(1,np,k,q),qmin(np,1,k,q),qmin(np,np,k,q))
           emin(k,q,ie)=max(emin(k,q,ie),0d0)
           emax(k,q,ie)=max(qmax(1,1,k,q),qmax(1,np,k,q),qmax(np,1,k,q),qmax(np,np,k,q))
        enddo
      enddo
   enddo
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
!$OMP BARRIER
#endif
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end subroutine

#endif







subroutine make_C0(zeta,elem,hybrid,nets,nete)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! apply DSS (aka assembly procedure) to zeta.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(in), target :: elem(:)
integer :: nets,nete
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: zeta

! local
integer :: k,i,j,ie,ic,kptr


call initEdgeBuffer(hybrid%par,edge1,elem,nlev)

do ie=nets,nete
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k)
#endif
   do k=1,nlev
      zeta(:,:,k,ie)=zeta(:,:,k,ie)*elem(ie)%spheremp(:,:)
   enddo
   kptr=0
   call edgeVpack(edge1, zeta(1,1,1,ie),nlev,kptr,ie)
enddo
call bndry_exchangeV(hybrid,edge1)
do ie=nets,nete
   kptr=0
   call edgeVunpack(edge1, zeta(1,1,1,ie),nlev,kptr,ie)
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k)
#endif
   do k=1,nlev
      zeta(:,:,k,ie)=zeta(:,:,k,ie)*elem(ie)%rspheremp(:,:)
   enddo
enddo
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
!$OMP BARRIER
#endif
#endif

call FreeEdgeBuffer(edge1)
end subroutine


subroutine make_C0_vector(v,elem,hybrid,nets,nete)
type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(in), target :: elem(:)
integer :: nets,nete
real (kind=real_kind), dimension(np,np,2,nlev,nets:nete) :: v

! local
integer :: k,i,j,ie,ic,kptr
type (EdgeBuffer_t)          :: edge2


call initEdgeBuffer(hybrid%par,edge2,elem,2*nlev)

do ie=nets,nete
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k)
#endif
   do k=1,nlev
      v(:,:,1,k,ie)=v(:,:,1,k,ie)*elem(ie)%spheremp(:,:)
      v(:,:,2,k,ie)=v(:,:,2,k,ie)*elem(ie)%spheremp(:,:)
   enddo
   kptr=0
   call edgeVpack(edge2, v(1,1,1,1,ie),2*nlev,kptr,ie)
enddo
call bndry_exchangeV(hybrid,edge2)
do ie=nets,nete
   kptr=0
   call edgeVunpack(edge2, v(1,1,1,1,ie),2*nlev,kptr,ie)
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k)
#endif
   do k=1,nlev
      v(:,:,1,k,ie)=v(:,:,1,k,ie)*elem(ie)%rspheremp(:,:)
      v(:,:,2,k,ie)=v(:,:,2,k,ie)*elem(ie)%rspheremp(:,:)
   enddo
enddo
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
!$OMP BARRIER
#endif
#endif

call FreeEdgeBuffer(edge2)
end subroutine






subroutine compute_zeta_C0_contra(zeta,elem,hybrid,nets,nete,nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute C0 vorticity.  That is, solve:
!     < PHI, zeta > = <PHI, curl(elem%state%v >
!
!    input:  v (stored in elem()%, in contra-variant coordinates)
!    output: zeta(:,:,:,:)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(in), target :: elem(:)
integer :: nt,nets,nete
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: zeta
real (kind=real_kind), dimension(np,np,2) :: ulatlon
real (kind=real_kind), dimension(np,np) :: v1,v2

! local
integer :: k,ie
type (derivative_t)          :: deriv

call derivinit(deriv)

do k=1,nlev
do ie=nets,nete
    v1 = elem(ie)%state%v(:,:,1,k,nt)
    v2 = elem(ie)%state%v(:,:,2,k,nt)
    ulatlon(:,:,1) = elem(ie)%D(:,:,1,1)*v1 + elem(ie)%D(:,:,1,2)*v2
    ulatlon(:,:,2) = elem(ie)%D(:,:,2,1)*v1 + elem(ie)%D(:,:,2,2)*v2
   !    zeta(:,:,k,ie)=elem(ie)%state%zeta(:,:,k)
   ! zeta(:,:,k,ie)=vorticity_sphere(ulatlon,deriv,elem(ie))
   call vorticity_sphere(ulatlon,deriv,elem(ie),zeta(:,:,k,ie))
enddo
enddo

call make_C0(zeta,elem,hybrid,nets,nete)

end subroutine



subroutine compute_div_C0_contra(zeta,elem,hybrid,nets,nete,nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute C0 divergence. That is, solve:
!     < PHI, zeta > = <PHI, div(elem%state%v >
!
!    input:  v (stored in elem()%, in contra-variant coordinates)
!    output: zeta(:,:,:,:)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(in), target :: elem(:)
integer :: nt,nets,nete
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: zeta
real (kind=real_kind), dimension(np,np,2) :: ulatlon
real (kind=real_kind), dimension(np,np) :: v1,v2

! local
integer :: k,ie
type (derivative_t)          :: deriv

call derivinit(deriv)

do k=1,nlev
do ie=nets,nete
    v1 = elem(ie)%state%v(:,:,1,k,nt)
    v2 = elem(ie)%state%v(:,:,2,k,nt)
    ulatlon(:,:,1) = elem(ie)%D(:,:,1,1)*v1 + elem(ie)%D(:,:,1,2)*v2
    ulatlon(:,:,2) = elem(ie)%D(:,:,2,1)*v1 + elem(ie)%D(:,:,2,2)*v2
   !    zeta(:,:,k,ie)=elem(ie)%state%zeta(:,:,k)
   ! zeta(:,:,k,ie)=divergence_sphere(ulatlon,deriv,elem(ie))
   call divergence_sphere(ulatlon,deriv,elem(ie),zeta(:,:,k,ie))
enddo
enddo

call make_C0(zeta,elem,hybrid,nets,nete)

end subroutine

subroutine compute_zeta_C0_par(zeta,elem,par,nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute C0 vorticity.  That is, solve:
!     < PHI, zeta > = <PHI, curl(elem%state%v >
!
!    input:  v (stored in elem()%, in lat-lon coordinates)
!    output: zeta(:,:,:,:)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
type (parallel_t) :: par
type (element_t)     , intent(in), target :: elem(:)
real (kind=real_kind), dimension(np,np,nlev,nelemd) :: zeta
integer :: nt

! local
type (hybrid_t)              :: hybrid
integer :: k,i,j,ie,ic
type (derivative_t)          :: deriv

! single thread
hybrid = config_thread_region(par,'serial')
!hybrid = hybrid_create(par,0,1)

call compute_zeta_C0_hybrid(zeta,elem,hybrid,1,nelemd,nt)

end subroutine


subroutine compute_div_C0_par(zeta,elem,par,nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute C0 divergence. That is, solve:
!     < PHI, zeta > = <PHI, div(elem%state%v >
!
!    input:  v (stored in elem()%, in lat-lon coordinates)
!    output: zeta(:,:,:,:)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

type (parallel_t) :: par
type (element_t)     , intent(in), target :: elem(:)
real (kind=real_kind), dimension(np,np,nlev,nelemd) :: zeta
integer :: nt

! local
type (hybrid_t)              :: hybrid
integer :: k,i,j,ie,ic
type (derivative_t)          :: deriv

! single thread
!hybrid = hybrid_create(par,0,1)
hybrid = config_thread_region(par,'serial')

call compute_div_C0_hybrid(zeta,elem,hybrid,1,nelemd,nt)

end subroutine



subroutine compute_zeta_C0_hybrid(zeta,elem,hybrid,nets,nete,nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute C0 vorticity.  That is, solve:
!     < PHI, zeta > = <PHI, curl(elem%state%v >
!
!    input:  v (stored in elem()%, in lat-lon coordinates)
!    output: zeta(:,:,:,:)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(in), target :: elem(:)
integer :: nt,nets,nete
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: zeta

! local
integer :: k,i,j,ie,ic
type (derivative_t)          :: deriv

call derivinit(deriv)

do ie=nets,nete
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k)
#endif
do k=1,nlev
   !    zeta(:,:,k,ie)=elem(ie)%state%zeta(:,:,k)
   !zeta(:,:,k,ie)=vorticity_sphere(elem(ie)%state%v(:,:,:,k,nt),deriv,elem(ie))
   call vorticity_sphere(elem(ie)%state%v(:,:,:,k,nt),deriv,elem(ie),zeta(:,:,k,ie))
enddo
enddo

call make_C0(zeta,elem,hybrid,nets,nete)

end subroutine


subroutine compute_div_C0_hybrid(zeta,elem,hybrid,nets,nete,nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! compute C0 divergence. That is, solve:
!     < PHI, zeta > = <PHI, div(elem%state%v >
!
!    input:  v (stored in elem()%, in lat-lon coordinates)
!    output: zeta(:,:,:,:)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(in), target :: elem(:)
integer :: nt,nets,nete
real (kind=real_kind), dimension(np,np,nlev,nets:nete) :: zeta

! local
integer :: k,i,j,ie,ic
type (derivative_t)          :: deriv

call derivinit(deriv)

do ie=nets,nete
#if (defined COLUMN_OPENMP)
!$omp parallel do private(k)
#endif
do k=1,nlev
   !    zeta(:,:,k,ie)=elem(ie)%state%zeta(:,:,k)
   ! zeta(:,:,k,ie)=divergence_sphere(elem(ie)%state%v(:,:,:,k,nt),deriv,elem(ie))
   call divergence_sphere(elem(ie)%state%v(:,:,:,k,nt),deriv,elem(ie),zeta(:,:,k,ie))
enddo
enddo

call make_C0(zeta,elem,hybrid,nets,nete)

end subroutine








#ifdef _PRIM
#define SW_NEIGHBOR_MINMAX
#ifdef SW_NEIGHBOR_MINMAX
subroutine neighbor_minmax(hybrid,edgeMinMax,nets,nete,min_neigh,max_neigh)
  !
  ! compute Q min&max over the element and all its neighbors
  !
  !
  use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
  type (hybrid_t)      , intent(in) :: hybrid
  type (EdgeBuffer_t)  , intent(inout) :: edgeMinMax
  integer :: nets, nete
  real (kind=real_kind) :: min_neigh(nlev,qsize,nets:nete)
  real (kind=real_kind) :: max_neigh(nlev,qsize,nets:nete)

  ! local
  integer :: ie,k,q
  integer :: kblk,qblk,kptr
  !real (kind=real_kind) :: Qmin(np,np,nlev,qsize)
  !real (kind=real_kind) :: Qmax(np,np,nlev,qsize)
  ! SUNWAY
  external :: slave_edgespack_es
  type param_edgeSpack_t
    integer*8 :: qmax, qmin, buf
    integer*8 :: putmap
    integer :: nets, nete, qsize
  end type param_edgeSpack_t
  type(param_edgeSpack_t) :: param_edgeSpack_s

  external :: slave_edgesunpack_es
  type param_edgeSunpack_t
    integer*8 :: qmax, qmin, receive
    integer*8 :: getmap
    integer :: nets, nete, qsize
  end type param_edgeSunpack_t
  type(param_edgeSunpack_t) :: param_edgeSunpack_s
  kblk = nlev  ! calculate size of the block of vertical levels
  qblk = qsize   ! calculate size of the block of tracers
  param_edgeSpack_s%qmin = loc(min_neigh)
  param_edgeSpack_s%qmax = loc(max_neigh)
  param_edgeSpack_s%buf = loc(edgeMinMax%buf)
  param_edgeSpack_s%putmap = loc(edgeMinMax%putmap(1,nets))
  param_edgeSpack_s%nets = nets
  param_edgeSpack_s%nete = nete
  param_edgeSpack_s%qsize = qsize

  param_edgeSunpack_s%qmax = loc(max_neigh)
  param_edgeSunpack_s%qmin = loc(min_neigh)
  param_edgeSunpack_s%receive = loc(edgeMinMax%receive)
  param_edgeSunpack_s%getmap = loc(edgeMinMax%getmap(1,nets))
  param_edgeSunpack_s%nets = nets
  param_edgeSunpack_s%nete = nete
  param_edgeSunpack_s%qsize = qsize

  call athread_spawn(slave_edgespack_es, param_edgeSpack_s)
  call athread_join()
  call t_startf('sw_biharmonic_minmax_bndry_exchangeV_vis')
  call bndry_exchangeS(hybrid,edgeMinMax,location='neighbor_minmax')
  call t_stopf('sw_biharmonic_minmax_bndry_exchangeV_vis')
  call athread_spawn(slave_edgesunpack_es, param_edgeSunpack_s)
  call athread_join()
end subroutine neighbor_minmax
#else
subroutine neighbor_minmax(hybrid,edgeMinMax,nets,nete,min_neigh,max_neigh)
!
! compute Q min&max over the element and all its neighbors
!
!
use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
type (hybrid_t)      , intent(in) :: hybrid
type (EdgeBuffer_t)  , intent(inout) :: edgeMinMax
integer :: nets, nete
real (kind=real_kind) :: min_neigh(nlev,qsize,nets:nete)
real (kind=real_kind) :: max_neigh(nlev,qsize,nets:nete)

! local
integer :: ie,k,q
integer :: kblk,qblk,kptr
!real (kind=real_kind) :: Qmin(np,np,nlev,qsize)
!real (kind=real_kind) :: Qmax(np,np,nlev,qsize)


   call t_startf('local_neighbor_minmax')
   kblk = nlev  ! calculate size of the block of vertical levels
   qblk = qsize   ! calculate size of the block of tracers

   do ie=nets,nete
      do q = 1, qsize
         kptr = nlev*(q - 1)
         call  edgeSpack(edgeMinMax,min_neigh(:,q,ie),kblk,kptr,ie)
         kptr = qsize*nlev + nlev*(q - 1)
         call  edgeSpack(edgeMinMax,max_neigh(:,q,ie),kblk,kptr,ie)
      enddo
   enddo

   call bndry_exchangeS(hybrid,edgeMinMax,location='neighbor_minmax')

   do ie=nets,nete
      do q=1,qsize
         kptr = nlev*(q - 1)
         call  edgeSunpackMIN(edgeMinMax,min_neigh(:,q,ie),kblk,kptr,ie)
         kptr = qsize*nlev + nlev*(q - 1)
         call  edgeSunpackMAX(edgeMinMax,max_neigh(:,q,ie),kblk,kptr,ie)
         do k=1,nlev
            min_neigh(k,q,ie) = max(min_neigh(k,q,ie),0d0)
         enddo
      enddo
   enddo
   call t_stopf('local_neighbor_minmax')
end subroutine neighbor_minmax
#endif


#define SW_NEIGHBOR_MINMAX_START
#ifdef SW_NEIGHBOR_MINMAX_START
subroutine neighbor_minmax_start(hybrid,edgeMinMax,nets,nete,min_neigh,max_neigh)
  use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
  type (hybrid_t)      , intent(in) :: hybrid
  type (EdgeBuffer_t)  , intent(inout) :: edgeMinMax
  integer :: nets,nete
  real (kind=real_kind) :: min_neigh(nlev,qsize,nets:nete)
  real (kind=real_kind) :: max_neigh(nlev,qsize,nets:nete)
  integer :: kblk, qblk
  integer :: kbeg, kend, qbeg, qend
  ! local
  integer :: ie,q, k,kptr
  external :: slave_edgespack_es
  type param_edgeSpack_t
    integer*8 :: qmax, qmin, buf
    integer*8 :: putmap
    integer :: nets, nete, qsize
  end type param_edgeSpack_t
  type(param_edgeSpack_t) :: param_edgeSpack_s

  !call get_loop_ranges(hybrid,kbeg=kbeg,kend=kend,qbeg=qbeg,qend=qend)
  kbeg = 1
  kend = nlev
  qbeg = 1
  qend = qsize
  kblk = kend - kbeg + 1   ! calculate size of the block of vertical levels
  qblk = qend - qbeg + 1   ! calculate size of the block of tracers
  call t_startf('sw_neighbor_minmax_start_vis')
  kblk = nlev  ! calculate size of the block of vertical levels
  qblk = qsize   ! calculate size of the block of tracers
  param_edgeSpack_s%qmin = loc(min_neigh)
  param_edgeSpack_s%qmax = loc(max_neigh)
  param_edgeSpack_s%buf = loc(edgeMinMax%buf)
  param_edgeSpack_s%putmap = loc(edgeMinMax%putmap(1,nets))
  param_edgeSpack_s%nets = nets
  param_edgeSpack_s%nete = nete
  param_edgeSpack_s%qsize = qsize
  call athread_spawn(slave_edgespack_es, param_edgeSpack_s)
  call athread_join()
  call t_startf('sw_neighbor_minmax_start_bndry_exchangeS_vis')
  call bndry_exchangeS_start(hybrid,edgeMinMax,location='viscosity_mod:882')
  call t_stopf('sw_neighbor_minmax_start_bndry_exchangeS_vis')
  call t_stopf('sw_neighbor_minmax_start_vis')
end subroutine neighbor_minmax_start
#else
subroutine neighbor_minmax_start(hybrid,edgeMinMax,nets,nete,min_neigh,max_neigh)
   use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
   type (hybrid_t)      , intent(in) :: hybrid
   type (EdgeBuffer_t)  , intent(inout) :: edgeMinMax
   integer :: nets,nete
   real (kind=real_kind) :: min_neigh(nlev,qsize,nets:nete)
   real (kind=real_kind) :: max_neigh(nlev,qsize,nets:nete)
   integer :: kblk, qblk
   integer :: kbeg, kend, qbeg, qend

   ! local
   integer :: ie,q, k,kptr


   call t_startf('local_neighbor_minmax_start')
   !call get_loop_ranges(hybrid,kbeg=kbeg,kend=kend,qbeg=qbeg,qend=qend)
   kbeg = 1
   kend = nlev
   qbeg = 1
   qend = qsize

   kblk = kend - kbeg + 1   ! calculate size of the block of vertical levels
   qblk = qend - qbeg + 1   ! calculate size of the block of tracers

   do ie=nets,nete
      do q=qbeg, qend
         kptr = nlev*(q - 1) + kbeg - 1
         call  edgeSpack(edgeMinMax,min_neigh(kbeg:kend,q,ie),kblk,kptr,ie)
         kptr = qsize*nlev + nlev*(q - 1) + kbeg - 1
         call  edgeSpack(edgeMinMax,max_neigh(kbeg:kend,q,ie),kblk,kptr,ie)
      enddo
   enddo

   call bndry_exchangeS_start(hybrid,edgeMinMax,location='viscosity_mod:882')
   call t_stopf('local_neighbor_minmax_start')
end subroutine neighbor_minmax_start
#endif


#define SW_NEIGHBOR_MINMAX_FINISH
#ifdef SW_NEIGHBOR_MINMAX_FINISH
subroutine neighbor_minmax_finish(hybrid,edgeMinMax,nets,nete,min_neigh,max_neigh)
  use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
  type (hybrid_t)      , intent(in) :: hybrid
  type (EdgeBuffer_t)  , intent(inout) :: edgeMinMax
  integer :: nets,nete
  real (kind=real_kind) :: min_neigh(nlev,qsize,nets:nete)
  real (kind=real_kind) :: max_neigh(nlev,qsize,nets:nete)
  integer :: kblk, qblk
  integer :: ie,q, k,kptr
  integer :: kbeg, kend, qbeg, qend
  external :: slave_edgesunpack_es
  type param_edgeSunpack_t
    integer*8 :: qmax, qmin, receive
    integer*8 :: getmap
    integer :: nets, nete, qsize
  end type param_edgeSunpack_t
  type(param_edgeSunpack_t) :: param_edgeSunpack_s
  !call get_loop_ranges(hybrid,kbeg=kbeg,kend=kend,qbeg=qbeg,qend=qend)
  qbeg=1
  qend=qsize
  kbeg=1
  kend=nlev
  kblk = kend - kbeg + 1   ! calculate size of the block of vertical levels
  qblk = qend - qbeg + 1   ! calculate size of the block of tracers
  param_edgeSunpack_s%qmax = loc(max_neigh)
  param_edgeSunpack_s%qmin = loc(min_neigh)
  param_edgeSunpack_s%receive = loc(edgeMinMax%receive)
  param_edgeSunpack_s%getmap = loc(edgeMinMax%getmap(1,nets))
  param_edgeSunpack_s%nets = nets
  param_edgeSunpack_s%nete = nete
  param_edgeSunpack_s%qsize = qsize
  call t_startf('sw_neighbor_minmax_finish_vis')
  call t_startf('sw_neighbor_minmax_finish_bndry_exchangeS_s')
  call bndry_exchangeS_finish(hybrid,edgeMinMax,location='viscosity_mod:902')
  call t_stopf('sw_neighbor_minmax_finish_bndry_exchangeS_s')
  call athread_spawn(slave_edgesunpack_es, param_edgeSunpack_s)
  call athread_join()
  call t_stopf('sw_neighbor_minmax_finish_vis')
end subroutine neighbor_minmax_finish
#else
subroutine neighbor_minmax_finish(hybrid,edgeMinMax,nets,nete,min_neigh,max_neigh)
   use perf_mod      , only : t_startf, t_stopf            ! _EXTERNAL
   type (hybrid_t)      , intent(in) :: hybrid
   type (EdgeBuffer_t)  , intent(inout) :: edgeMinMax
   integer :: nets,nete
   real (kind=real_kind) :: min_neigh(nlev,qsize,nets:nete)
   real (kind=real_kind) :: max_neigh(nlev,qsize,nets:nete)
   integer :: kblk, qblk
   integer :: ie,q, k,kptr
   integer :: kbeg, kend, qbeg, qend

   !call get_loop_ranges(hybrid,kbeg=kbeg,kend=kend,qbeg=qbeg,qend=qend)
   qbeg=1
   qend=qsize
   kbeg=1
   kend=nlev

   kblk = kend - kbeg + 1   ! calculate size of the block of vertical levels
   qblk = qend - qbeg + 1   ! calculate size of the block of tracers

   call t_startf('local_neighbor_minmax_finish')
   call bndry_exchangeS_finish(hybrid,edgeMinMax,location='viscosity_mod:902')

   do ie=nets,nete
      do q=qbeg, qend
         kptr = nlev*(q - 1) + kbeg - 1
         call  edgeSunpackMIN(edgeMinMax,min_neigh(kbeg:kend,q,ie),kblk,kptr,ie)
         kptr = qsize*nlev + nlev*(q - 1) + kbeg - 1
         call  edgeSunpackMAX(edgeMinMax,max_neigh(kbeg:kend,q,ie),kblk,kptr,ie)
         do k=kbeg,kend
            min_neigh(k,q,ie) = max(min_neigh(k,q,ie),0d0)
         enddo
      enddo
   enddo
   call t_stopf('local_neighbor_minmax_finish')
end subroutine neighbor_minmax_finish
#endif



#else



subroutine neighbor_minmax(elem,hybrid,edgeMinMax,nets,nete,nt,min_neigh,max_neigh,min_var,max_var,kmass)
!
! compute Q min&max over the element and all its neighbors
!
!
integer :: nets,nete,nt
type (hybrid_t)      , intent(in) :: hybrid
type (element_t)     , intent(inout) :: elem(:)
type (EdgeBuffer_t)  , intent(in) :: edgeMinMax
real (kind=real_kind) :: min_neigh(nlev,nets:nete)
real (kind=real_kind) :: max_neigh(nlev,nets:nete)
real (kind=real_kind),optional :: min_var(nlev,nets:nete)
real (kind=real_kind),optional :: max_var(nlev,nets:nete)
real (kind=real_kind) :: Qmin(np,np,nlev)
real (kind=real_kind) :: Qmax(np,np,nlev)
real (kind=real_kind) :: Qvar(np,np,nlev)
type (EdgeBuffer_t)          :: edgebuf
integer, optional :: kmass
type (EdgeDescriptor_t), allocatable :: desc(:)

! local
integer :: ie,k,q

  if(present(kmass))then
!the check if kmass is a valid number is done in sweq_mod
    do k=1,nlev
      if(k.ne.kmass)then
         do ie=nets,nete
            elem(ie)%state%p(:,:,k,nt)=elem(ie)%state%p(:,:,k,nt)/&
            elem(ie)%state%p(:,:,kmass,nt)
         enddo
      endif
    enddo
  endif

    ! create edge buffer for 3 fields
    call initEdgeBuffer(hybrid%par,edgebuf,elem,3*nlev)


    ! compute p min, max
    do ie=nets,nete
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
       do k=1,nlev
          Qmin(:,:,k)=minval(elem(ie)%state%p(:,:,k,nt))
          Qmax(:,:,k)=maxval(elem(ie)%state%p(:,:,k,nt))
          ! max - min - crude approximation to TV within the element:
          Qvar(:,:,k)=Qmax(1,1,k)-Qmin(1,1,k)
       enddo
       call edgeVpack(edgebuf,Qmax,nlev,0,ie)
       call edgeVpack(edgebuf,Qmin,nlev,nlev,ie)
       call edgeVpack(edgebuf,Qvar,nlev,2*nlev,ie)
    enddo

    call bndry_exchangeV(hybrid,edgebuf,location='viscosity_mod:976')


    do ie=nets,nete
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
       do k=1,nlev
          Qmin(:,:,k)=minval(elem(ie)%state%p(:,:,k,nt))
          Qmax(:,:,k)=maxval(elem(ie)%state%p(:,:,k,nt))
       enddo

       ! now unpack the min
       if (present(min_var)) then
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
          do k=1,nlev
             Qvar(:,:,k)=Qmax(1,1,k)-Qmin(1,1,k)
          enddo
! WARNING - edgeVunpackMin/Max take second argument as input/ouput
          call edgeVunpackMin(edgebuf,Qvar,nlev,2*nlev,ie)
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
          do k=1,nlev
             min_var(k,ie)=minval(Qvar(:,:,k))
          enddo
       endif

       ! now unpack the max
       if (present(max_var)) then
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
          do k=1,nlev
             Qvar(:,:,k)=Qmax(1,1,k)-Qmin(1,1,k)
          enddo
! WARNING - edgeVunpackMin/Max take second argument as input/ouput
          call edgeVunpackMax(edgebuf,Qvar,nlev,2*nlev,ie)
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
          do k=1,nlev
             max_var(k,ie)=maxval(Qvar(:,:,k))
          enddo
       endif


! WARNING - edgeVunpackMin/Max take second argument as input/ouput
       call edgeVunpackMax(edgebuf,Qmax,nlev,0,ie)
       call edgeVunpackMin(edgebuf,Qmin,nlev,nlev,ie)
#if (defined _LOOP_OPENMP)
!$omp parallel do num_threads(vert_num_threads)
#endif
       do k=1,nlev
          max_neigh(k,ie)=maxval(Qmax(:,:,k))
          min_neigh(k,ie)=minval(Qmin(:,:,k))
       enddo

    end do

    call FreeEdgeBuffer(edgebuf)
#ifdef DEBUGOMP
!$OMP BARRIER
#endif

  if(present(kmass))then
    do k=1,nlev
       if(k.ne.kmass)then
          do ie=nets,nete
             elem(ie)%state%p(:,:,k,nt)=elem(ie)%state%p(:,:,k,nt)*&
             elem(ie)%state%p(:,:,kmass,nt)
          enddo
       endif
    enddo
  endif
end subroutine


#endif



  subroutine test_ibyp(elem, hybrid,  nets,   nete)
!
! Note: vector test functions should be co-variant since u is contra-variant
!  PHIvec = PHIcov  (test function)
!  PHIcon = DtD PHIcov
!
! weak grad:
!  < PHIcov du/dt > = < PHIcon grad(p) >    (output of grad is covariant)
!  < PHIcov du/dt > = -< div(PHIcon) p >    (input of div is contra)
!  verify:
!    gradient_sphere_wk(p) = - <div(PHIcon) p >
!    gradient_sphere_wk(p) + MASS*grad(p) = b.c. (b.c. are covariant)
!
! weak div:
!   < PHI div(u) > = -< grad(PHI) dot u >     u=contra, output of grad is covariant
! verify:
!   divergence_sphere_wk(u) = -<grad(PHI) dot u>
!   divergence_sphere_wk(u) + MASS*div(u) = b.c.  (b.c. are scalars)
!
! weak curl:
!  < PHIcov du/dt > = < PHIcov curl( a ) >    (output of curl is contra)
!  < PHIcov du/dt > = < vor(PHIcov) a >       (input to vor is covariant)
! verify:
!    curl_sphere_wk(a) = < vor(PHIcov) a >
!    curl_sphere_wk(a) - MASS*curl(a) = b.c. (b.c. are contra)
!
    ! ---------------------
    use kinds, only : real_kind
    ! ---------------------
    use physical_constants, only : rearth
    ! ---------------------
    use dimensions_mod, only : np, nlev
    ! ---------------------
    use element_mod, only : element_t
    ! ---------------------
    use hybrid_mod, only : hybrid_t
    ! ---------------------
    use derivative_mod, only : derivative_t, gradient_sphere, divergence_sphere,vorticity_sphere,&
                               divergence_sphere_wk, curl_sphere
    use global_norms_mod

    implicit none

    type (element_t)     , intent(inout), target :: elem(:)

    type (hybrid_t)      , intent(in) :: hybrid

    integer              , intent(in) :: nets
    integer              , intent(in) :: nete

#if 0
#undef CURLGRAD_TEST
#define IBYP_TEST
    ! =================
    ! Local
    ! =================
    ! pointer ...
    real (kind=real_kind), dimension(:,:), pointer :: rspheremv,spheremv

    ! Thread private working set ...

    real (kind=real_kind), dimension(np,np,nets:nete) :: ptens
    real (kind=real_kind), dimension(np,np,nets:nete) :: ptens2
    real (kind=real_kind), dimension(np,np,nets:nete) :: ptens3

    real (kind=real_kind), dimension(np,np,2,nets:nete)    :: pv      ! p*v lat-lon
    real (kind=real_kind), dimension(np,np,nets:nete)            :: E          ! kinetic energy term
    real (kind=real_kind), dimension(np,np,nets:nete)  :: divbig
    real (kind=real_kind), dimension(np,np,nets:nete)  :: wdivbig
    real (kind=real_kind), dimension(np,np,2,nets:nete)    :: gradbig

    real (kind=real_kind), dimension(np,np,2)      :: ulatlon
    real (kind=real_kind), dimension(np,np,2)      :: grade
    real (kind=real_kind), dimension(np,np,2)      :: grade2
    real (kind=real_kind), dimension(np,np)      :: vor
    real (kind=real_kind), dimension(np,np)      :: div
    real (kind=real_kind), dimension(np,np)      :: wdiv

    real (kind=real_kind) ::  v1,v2,v3

    real*8                :: st,et, time_adv
    integer    :: i,j,k,ie,iie,ii,jj
    integer    :: kptr
    integer    :: nm1,n0,np1
    integer    :: nstep

    type (derivative_t)          :: deriv
    call derivinit(deriv)



    ! ===================================
    ! construct test function
    ! ===================================

    do iie=nets,nete
       do jj=1,np
          do ii=1,np
             ! test for cardinal function at iie,jj,ii

    write(iulog,'(a,3i4,2e20.10)') 'carinal function:  ',iie,ii,jj

    ! construct two global cardinal functions  pv and E
    do ie=nets,nete
       do j=1,np
          do i=1,np


             E(i,j,ie)=0
#ifdef CURLGRAD_TEST
             if (ie==iie .and. i==ii .and. j==jj) E(i,j,ie)=1
#else
             if (ie==1 .and. i==1 .and. j==1) E(i,j,ie)=1
             if (ie==1 .and. i==1 .and. j==2) E(i,j,ie)=1
             if (ie==1 .and. i==2 .and. j==1) E(i,j,ie)=1
             if (ie==1 .and. i==3 .and. j==3) E(i,j,ie)=1
#endif

             ! delta function in contra coordinates
             v1     = 0
             v2     = 0
             if (ie==iie .and. i==ii .and. j==jj) then
                !v1=rearth
                v2=rearth
             endif

             ulatlon(i,j,1)=elem(ie)%D(1,1,i,j)*v1 + elem(ie)%D(1,2,i,j)*v2   ! contra->latlon
             ulatlon(i,j,2)=elem(ie)%D(2,1,i,j)*v1 + elem(ie)%D(2,2,i,j)*v2   ! contra->latlon
             pv(i,j,1,ie) = ulatlon(i,j,1)
             pv(i,j,2,ie) = ulatlon(i,j,2)

          end do
       end do
    enddo
    call make_C0(E,elem,hybrid,nets,nete)
    call make_C0_vector(pv,elem,hybrid,nets,nete)


#ifdef CURLGRAD_TEST
    ! check curl(grad(E))
    do ie=nets,nete
       if ( maxval(abs(E(:,:,ie))) > 0 ) then
       !write(iulog,'(a,i4,2e20.10)') 'maxval: E =',ie,maxval(E(:,:,ie))
       ! grade=curl_sphere(E(:,:,ie),deriv,elem(ie))
       call curl_sphere(E(:,:,ie),deriv,elem(ie),grade)
       call divergence_sphere(grade,deriv,elem(ie),div)
       call vorticity_sphere(grade,deriv,elem(ie),vor)
       if (maxval(abs(div))*rearth**2 > .2e-11) then
          write(iulog,'(a,i4,2e20.10)') 'maxval: div(curl),  vor(curl)=',ie,maxval(abs(div))*rearth**2,maxval(abs(vor))*rearth**2
       endif

       call gradient_sphere(E(:,:,ie),deriv,elem(ie)%Dinv,grade)
       call vorticity_sphere(grade,deriv,elem(ie),vor)
       call divergence_sphere(grade,deriv,elem(ie),div)
       if (maxval(abs(vor))*rearth**2 > .2e-11) then
          write(iulog,'(a,i4,2e20.10)') 'maxval: curl(grad), div(grad)=',ie,maxval(abs(vor))*rearth**2,maxval(abs(div))*rearth**2
       endif
       endif
    enddo

    ! check div(curl(E)) with DSS
    do ie=nets,nete
       ! gradbig(:,:,:,ie)=curl_sphere(E(:,:,ie),deriv,elem(ie))
       call curl_sphere(E(:,:,ie),deriv,elem(ie),gradbig(:,:,:,ie))
    enddo
    call make_C0_vector(gradbig,elem,hybrid,nets,nete)
    do ie=nets,nete
       ! divbig(:,:,ie)=divergence_sphere(gradbig(:,:,:,ie),deriv,elem(ie))
       call divergence_sphere(gradbig(:,:,:,ie),deriv,elem(ie),divbig(:,:,ie))
    enddo
    call make_C0(divbig,elem,hybrid,nets,nete)
    do ie=nets,nete
       if (maxval(abs(divbig(:,:,ie)))*rearth**2 > .8e-12) then
          write(iulog,'(a,i4,2e20.10)') 'maxval: [div([curl])]=',ie,maxval(abs(divbig(:,:,ie)))*rearth**2
       endif
    enddo


    ! check curl(grad(E)) with DSS
    do ie=nets,nete
       ! gradbig(:,:,:,ie)=gradient_sphere(E(:,:,ie),deriv,elem(ie)%Dinv)
       call gradient_sphere(E(:,:,ie),deriv,elem(ie)%Dinv,gradbig(:,:,:,ie))
    enddo
    call make_C0_vector(gradbig,elem,hybrid,nets,nete)
    do ie=nets,nete
       ! divbig(:,:,ie)=vorticity_sphere(gradbig(:,:,:,ie),deriv,elem(ie))
       call vorticity_sphere(gradbig(:,:,:,ie),deriv,elem(ie),divbig(:,:,ie))
    enddo
    call make_C0(divbig,elem,hybrid,nets,nete)

    do ie=nets,nete
       if (maxval(abs(divbig(:,:,ie)))*rearth**2 > .8e-12) then
          write(iulog,'(a,i4,2e20.10)') 'maxval: [curl([gradl])]=',ie,maxval(abs(divbig(:,:,ie)))*rearth**2
       endif
    enddo
#endif


#ifdef IBYP_TEST
    ! compare <grad(E) dot pv> and <E div(pv)>  < E weak_div(pv) >
    v2=0
    do ie=nets,nete
       spheremv     => elem(ie)%spheremp(:,:)

       ! div = divergence_sphere(pv(1,1,1,ie),deriv,elem(ie))      ! latlon vector -> scalar
       call divergence_sphere(pv(1,1,1,ie),deriv,elem(ie),div)      ! latlon vector -> scalar
       !grade = gradient_sphere(E(1,1,ie),deriv,elem(ie)%Dinv)
       call gradient_sphere(E(1,1,ie),deriv,elem(ie)%Dinv,grade)
       !wdiv = divergence_sphere_wk(pv(1,1,1,ie),deriv,elem(ie))
       call divergence_sphere_wk(pv(1,1,1,ie),deriv,elem(ie),wdiv)



       do j=1,np
          do i=1,np
!             write(iulog,'(3i3,3e22.14)') ie,i,j,pv(i,j,1,ie),pv(i,j,2,ie),div(i,j)

             ! (grad(E) dot pv )
             ptens3(i,j,ie) = grade(i,j,1)*pv(i,j,1,ie) + grade(i,j,2)*pv(i,j,2,ie)
             v2=v2+wdiv(i,j)*E(i,j,ie)
          end do
       end do
       ptens(:,:,ie)=div(:,:)*E(:,:,ie)   ! < E div(pv) >
       ptens2(:,:,ie)=wdiv(:,:)*E(:,:,ie)/spheremv(:,:)   ! < wdiv E >
       ! ===================================================
       ! Pack cube edges of tendencies, rotate velocities
       ! ===================================================
       divbig(:,:,ie)=div(:,:)*spheremv(:,:)
       wdivbig(:,:,ie)=wdiv(:,:)
    end do
    call make_C0(divbig,elem,hybrid,nets,nete)
    call make_C0(wdivbig,elem,hybrid,nets,nete)


!!$    v1=global_integral(elem,ptens,hybrid,np,nets,nete)
!!$    v3=global_integral(elem,ptens3,hybrid,np,nets,nete)
!!$    print *,'< E div(pv) >   =',v1
!!$    print *,'< E div_wk(pv) >=',v2/(4*4*atan(1d0))
!!$    v2=global_integral(elem,ptens2,hybrid,np,nets,nete)
!!$    print *,'< E div_wk(pv) >=',v2
!!$    print *,'-<grad(E),pv >  =',-v3
!!$!    print *,'sum1-sum2/max',(v1-v2)/max(abs(v1),abs(v2))
!!$


    do ie=nets,nete
       div(:,:)=divbig(:,:,ie)
       wdiv(:,:)=wdivbig(:,:,ie)
       ! ===========================================================
       ! Compute velocity and pressure tendencies for all levels
       ! ===========================================================
          do j=1,np
             do i=1,np
                ! < div(pv) >   vs < div_wk(pv) >
                if ( abs(div(i,j)-wdiv(i,j)) > .15e-17) then
!                   write(iulog,'(3i3,4e22.14)') ie,i,j,div(i,j),wdiv(i,j),div(i,j)-wdiv(i,j),E(i,j,ie)
                endif
                if ( E(i,j,ie)/=0 ) then
!                   write(iulog,'(3i3,4e22.14)') ie,i,j,div(i,j),wdiv(i,j),div(i,j)-wdiv(i,j),E(i,j,ie)
                endif

             end do
          end do


    end do
#ifdef DEBUGOMP
#if (defined HORIZ_OPENMP)
!$OMP BARRIER
#endif
#endif
    write(iulog,'(a,3i4,2e20.10)') 'max diff div-wdiv: ',iie,ii,jj,maxval(abs(divbig(:,:,:)-wdivbig(:,:,:))),maxval(divbig(:,:,:))
#endif


    enddo
    enddo
    enddo
    stop
#endif
  end subroutine test_ibyp









  subroutine check_edge_flux(elem,deriv,nets,nete)
!
!  check local element vector dentities:
!*****
!  1. div and weak div are adjoints: (for all scalar test functions)
!     integral[  p div(u) ] + integral[ grad(p) dot u ] = boundary_integral[ p u dot n]
!       PHI div(u) spheremp - div_wk(u)(i,j) = boundary_integral[ u PHI]
!       where PHI = the delta function at (i,j)
!
!*****
!  2. grad and weak grad are adjoints:
!     weak gradient is defined with CONTRA vector test functions
!     i.e. it returns vector:   [  integral[ p div(PHIcontra_1) ]
!                               [  integral[ p div(PHIcontra_2) ]
!
!   integral[  p div(u) ] + integral[ grad(p) dot u ] = boundary_integral[ p u dot n]
! take u = PHIcontra_1 = (1,0) vector delta funciton at (i,j):
!  -grad_wk(p)_1(i,j) + spheremp PHIcontra_1 dot grad(p) = boundary_integral[ PHIcontra_1 p]
! and then take u = PHIcontra_2 = (0,1) vector delta function at (i,j):
!  -grad_wk(p)_2(i,j) + spheremp PHIcontra_2 dot grad(p) = boundary_integral[ PHIcontra_2 p]
!
! which is an equation for each covariant component:
! -grad_wk(p)_cov1 + spheremp grad(p)_cov1 = boundary_integral[ PHIcontra_1 p dot n]
! -grad_wk(p)_cov2 + spheremp grad(p)_cov2 = boundary_integral[ PHIcontra_2 p dot n]
!
! HOMME-SE works in latlon, so convert cov->lat/lon:
!
! -grad_wk(p) + spheremp grad(p) = D^-t * B
!
! with
!    B1 = boundary_integral[ PHIcontra_1 p]
!    B2 = boundary_integral[ PHIcontra_2 p]
!
!*****
! 3.  weak grid with COVARIANT test functions!
!   integral[  p div(u) ] + integral[ grad(p) dot u ] = boundary_integral[ p u dot n]
! take u = PHIcov_1 = (1,0) vector delta funciton at (i,j):
!  -grad_wk(p)_1(i,j) + spheremp PHIcov_1 dot grad(p) = boundary_integral[ PHIcov_1 p]
! and then take u = PHIcov_2 = (0,1) vector delta function at (i,j):
!  -grad_wk(p)_2(i,j) + spheremp PHIcov_2 dot grad(p) = boundary_integral[ PHIcov_2 p]
!
! which is an equation for each CONTRA component:
! -grad_wk(p)_contra1 + spheremp grad(p)_contra1 = B1
! -grad_wk(p)_contra2 + spheremp grad(p)_contra2 = B2
!
! HOMME-SE works in latlon, so convert contra ->lat/lon:
!
! -grad_wk(p) + spheremp grad(p) = D * B
!
! with
!    B1 = boundary_integral[ PHIcov_1 p]
!    B2 = boundary_integral[ PHIcov_2 p]
!
!*****
! 4.  weak curl with COVARIANT test functions!
!  integral[ u dot curl(v)] - integral[v dot curl(u)] = boundary_integral[ v cross u dot n]
!  curl(p) = curl(p*khat) = horizontal vector
!  vor(U) =  s*khat       = (which we treat as a scalar)
!   integral[ p * vor(u)  ] - integral[ u dot curl(p) ] = boundary_integral[ u cross p*khat  dot n]
!
! take u = PHIcov_1 = (1,0) vector delta funciton at (i,j):
!   curl_wk(p)_1(i,j) - spheremp PHIcov_1 dot curl(p) = boundary_integral[ perp(PHIcov_1) p]
! and then take u = PHIcov_2 = (0,1) vector delta function at (i,j):
!   curl_wk(p)_2(i,j) - spheremp PHIcov_2 dot curl(p) = boundary_integral[ perp(PHIcov_2) p]
!
! which is an equation for each CONTRA component:
! curl_wk(p)_contra1 - spheremp curl(p)_contra1 = B1
! curl_wk(p)_contra2 - spheremp curl(p)_contra2 = B2
!
! HOMME-SE works in latlon, so convert contra ->lat/lon:
!
! curl_wk(p) + spheremp curl(p) = D * B
!
! with
!    B1 = boundary_integral[ PHIcov_1 p]
!    B2 = boundary_integral[ PHIcov_2 p]
!
  use dimensions_mod, only : np, np, nlev
  use element_mod, only    : element_t
  use derivative_mod, only  : derivative_t, divergence_sphere, divergence_sphere_wk, &
                             element_boundary_integral, gradient_sphere, &
                             gradient_sphere_wk_testcontra,gradient_sphere_wk_testcov, &
                             curl_sphere, curl_sphere_wk_testcov
  use physical_constants, only : rrearth

  implicit none

  type (element_t)     , intent(inout), target :: elem(:)
  type (derivative_t)  , intent(in) :: deriv
  integer :: nets,nete
  ! local
  real (kind=real_kind), dimension(np,np,2) :: ucontra,ulatlon,gradp,gradp_wk,ucov
  real (kind=real_kind), dimension(np,np) :: phidivu,ugradphi,rhs,lhs,p
  real (kind=real_kind), dimension(np,np) :: rhs2,lhs2
  integer :: i,j,ie

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  print *,'integration by parts identity: check div/weak div:'
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! test integration by parts identity for each Cardinal function PHI:
  ! div(u)*spheremp - div_wk(u) = boundary integral phi u dot n
  do ie=nets,nete
     call random_number(ucontra)
     ! contra->latlon
     ulatlon(:,:,1)=(elem(ie)%D(:,:,1,1)*ucontra(:,:,1) + elem(ie)%D(:,:,1,2)*ucontra(:,:,2))
     ulatlon(:,:,2)=(elem(ie)%D(:,:,2,1)*ucontra(:,:,1) + elem(ie)%D(:,:,2,2)*ucontra(:,:,2))
     ! phidivu = elem(ie)%spheremp(:,:)*divergence_sphere(ulatlon,deriv,elem(ie))
     call divergence_sphere(ulatlon,deriv,elem(ie),phidivu)
     phidivu = phidivu*elem(ie)%spheremp
     ! ugradphi = divergence_sphere_wk(ulatlon,deriv,elem(ie))
     call divergence_sphere_wk(ulatlon,deriv,elem(ie),ugradphi)
     lhs = phidivu - ugradphi

     rhs = element_boundary_integral(ulatlon,deriv,elem(ie))


     do j=1,np
        do i=1,np
           if ( abs(lhs(i,j)-rhs(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: div/div_wk integration by parts failure!'
              write(*,'(a,2i3,a,3e12.5)') 'for test function (i,j)=',i,j,' lhs,rhs=',lhs(i,j),rhs(i,j),lhs(i,j)-rhs(i,j)
           endif
        enddo
     enddo
  enddo

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  print *,'check grad/weak grad (gradient_sphere_wk_testcontra)'
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! PHIVEC = contra cardinal function
  !          check each contra component seperately

  do ie=nets,nete
     call random_number(p)

     ! grad(p)  (lat/lon vector)
     !gradp = gradient_sphere(p,deriv,elem(ie)%Dinv)
     call gradient_sphere(p,deriv,elem(ie)%Dinv,gradp)
     gradp(:,:,1)=gradp(:,:,1)*elem(ie)%spheremp(:,:)
     gradp(:,:,2)=gradp(:,:,2)*elem(ie)%spheremp(:,:)
     !gradp_wk = gradient_sphere_wk_testcontra(p,deriv,elem(ie))
     call gradient_sphere_wk_testcontra(p,deriv,elem(ie),gradp_wk)

     ucontra(:,:,1)=p  ! PHIvec_1 * p
     ucontra(:,:,2)=0
     ! contra->latlon
     ulatlon(:,:,1)=(elem(ie)%D(:,:,1,1)*ucontra(:,:,1) + elem(ie)%D(:,:,1,2)*ucontra(:,:,2))
     ulatlon(:,:,2)=(elem(ie)%D(:,:,2,1)*ucontra(:,:,1) + elem(ie)%D(:,:,2,2)*ucontra(:,:,2))

     rhs = element_boundary_integral(ulatlon,deriv,elem(ie))
     lhs = gradp(:,:,1)-gradp_wk(:,:,1)

     ucontra(:,:,1)=0  ! PHIvec_2 * p
     ucontra(:,:,2)=p
     ! contra->latlon
     ulatlon(:,:,1)=(elem(ie)%D(:,:,1,1)*ucontra(:,:,1) + elem(ie)%D(:,:,1,2)*ucontra(:,:,2))
     ulatlon(:,:,2)=(elem(ie)%D(:,:,2,1)*ucontra(:,:,1) + elem(ie)%D(:,:,2,2)*ucontra(:,:,2))
     rhs2 = element_boundary_integral(ulatlon,deriv,elem(ie))
     lhs2 = gradp(:,:,2)-gradp_wk(:,:,2)


     ! boundary integral gives covariant components. (see above) convert to latlon:
     ! cov -> latlon
     gradp(:,:,1)=rhs
     gradp(:,:,2)=rhs2
     rhs(:,:)=elem(ie)%Dinv(:,:,1,1)*gradp(:,:,1) + elem(ie)%Dinv(:,:,2,1)*gradp(:,:,2)
     rhs2(:,:)=elem(ie)%Dinv(:,:,1,2)*gradp(:,:,1) + elem(ie)%Dinv(:,:,2,2)*gradp(:,:,2)


     do j=1,np
        do i=1,np
           if ( abs(lhs(i,j)-rhs(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: grad/grad_wk CONTRA (1) integration by parts failure!'
              write(*,'(a,2i3,a,4e12.4)') '(i,j)=',i,j,' lhs,rhs=',lhs(i,j),rhs(i,j),&
                   lhs(i,j)-rhs(i,j),lhs(i,j)/rhs(i,j)
           endif
        enddo
     enddo


     do j=1,np
        do i=1,np
           if ( abs(lhs2(i,j)-rhs2(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: grad/grad_wk CONTRA (2) integration by parts failure!'
              write(*,'(a,2i2,a,3e12.4)') '(i,j)=',i,j,' lhs,rhs=',lhs2(i,j),rhs2(i,j),lhs2(i,j)-rhs2(i,j)
           endif
        enddo
     enddo
  enddo

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  print *,'check grad/weak grad (gradient_sphere_wk_testcov)'
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  do ie=nets,nete
     call random_number(p)


     ! grad(p)  (lat/lon vector)
     !gradp = gradient_sphere(p,deriv,elem(ie)%Dinv)
     call gradient_sphere(p,deriv,elem(ie)%Dinv,gradp)
     gradp(:,:,1)=gradp(:,:,1)*elem(ie)%spheremp(:,:)
     gradp(:,:,2)=gradp(:,:,2)*elem(ie)%spheremp(:,:)
     ! gradp_wk = gradient_sphere_wk_testcov(p,deriv,elem(ie))
     call gradient_sphere_wk_testcov(p,deriv,elem(ie),gradp_wk)
     lhs = gradp(:,:,1)-gradp_wk(:,:,1)
     lhs2 = gradp(:,:,2)-gradp_wk(:,:,2)

     ucov(:,:,1)=p  ! PHIvec_1 * p
     ucov(:,:,2)=0
     ! cov->latlon
     ulatlon(:,:,1)=(elem(ie)%Dinv(:,:,1,1)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,1)*ucov(:,:,2))
     ulatlon(:,:,2)=(elem(ie)%Dinv(:,:,1,2)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,2)*ucov(:,:,2))
     rhs = element_boundary_integral(ulatlon,deriv,elem(ie))

     ucov(:,:,1)=0  ! PHIvec_2 * p
     ucov(:,:,2)=p
     ! cov->latlon
     ulatlon(:,:,1)=(elem(ie)%Dinv(:,:,1,1)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,1)*ucov(:,:,2))
     ulatlon(:,:,2)=(elem(ie)%Dinv(:,:,1,2)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,2)*ucov(:,:,2))
     rhs2 = element_boundary_integral(ulatlon,deriv,elem(ie))


     ! boundary integral gives contra components. (see above) convert to latlon:
     ! contra -> latlon
     gradp(:,:,1)=rhs
     gradp(:,:,2)=rhs2
     rhs(:,:) =elem(ie)%D(:,:,1,1)*gradp(:,:,1) + elem(ie)%D(:,:,1,2)*gradp(:,:,2)
     rhs2(:,:)=elem(ie)%D(:,:,2,1)*gradp(:,:,1) + elem(ie)%D(:,:,2,2)*gradp(:,:,2)


     do j=1,np
        do i=1,np
           if ( abs(lhs(i,j)-rhs(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: grad/grad_wk COV (1) integration by parts failure!'
              write(*,'(a,2i2,a,4e12.4)') '(i,j)=',i,j,' lhs,rhs=',lhs(i,j),rhs(i,j),lhs(i,j)-rhs(i,j),lhs(i,j)/rhs(i,j)
           endif
        enddo
     enddo

     do j=1,np
        do i=1,np
           if ( abs(lhs2(i,j)-rhs2(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: grad/grad_wk COV (2) integration by parts failure!'
              write(*,'(a,2i2,a,3e12.4)') '(i,j)=',i,j,' lhs,rhs=',lhs2(i,j),rhs2(i,j),lhs2(i,j)-rhs2(i,j)
           endif
        enddo
     enddo
  enddo

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  print *,'check curl/weak curl (curl_sphere_wk_testcov)'
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  do ie=nets,nete
     call random_number(p)

     ! grad(p)  (lat/lon vector)
     ! gradp = curl_sphere(p,deriv,elem(ie))
     call curl_sphere(p,deriv,elem(ie),gradp)
     gradp(:,:,1)=gradp(:,:,1)*elem(ie)%spheremp(:,:)
     gradp(:,:,2)=gradp(:,:,2)*elem(ie)%spheremp(:,:)
     ! gradp_wk = curl_sphere_wk_testcov(p,deriv,elem(ie))
     call curl_sphere_wk_testcov(p,deriv,elem(ie),gradp_wk)
     lhs =  gradp_wk(:,:,1)-gradp(:,:,1)
     lhs2 = gradp_wk(:,:,2)-gradp(:,:,2)

     ucov(:,:,1)=p  ! PHIvec_1 * p
     ucov(:,:,2)=0
     ! cov->latlon, and then u cross khat:
     ulatlon(:,:,2)=-(elem(ie)%Dinv(:,:,1,1)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,1)*ucov(:,:,2))
     ulatlon(:,:,1)= (elem(ie)%Dinv(:,:,1,2)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,2)*ucov(:,:,2))
     rhs = element_boundary_integral(ulatlon,deriv,elem(ie))

     ucov(:,:,1)=0  ! PHIvec_2 * p
     ucov(:,:,2)=p
     ! cov->latlon, and u cross khat:
     ulatlon(:,:,2)=-(elem(ie)%Dinv(:,:,1,1)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,1)*ucov(:,:,2))
     ulatlon(:,:,1)= (elem(ie)%Dinv(:,:,1,2)*ucov(:,:,1) + elem(ie)%Dinv(:,:,2,2)*ucov(:,:,2))
     rhs2 = element_boundary_integral(ulatlon,deriv,elem(ie))


     ! boundary integral gives contra components. (see above) convert to latlon:
     ! contra -> latlon
     gradp(:,:,1)=rhs
     gradp(:,:,2)=rhs2
     rhs(:,:) =elem(ie)%D(:,:,1,1)*gradp(:,:,1) + elem(ie)%D(:,:,1,2)*gradp(:,:,2)
     rhs2(:,:)=elem(ie)%D(:,:,2,1)*gradp(:,:,1) + elem(ie)%D(:,:,2,2)*gradp(:,:,2)


     do j=1,np
        do i=1,np
           if ( abs(lhs(i,j)-rhs(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: curl/curl_wk COV (1) integration by parts failure!'
              write(*,'(a,2i2,a,4e12.4)') '(i,j)=',i,j,' lhs,rhs=',lhs(i,j),rhs(i,j),lhs(i,j)-rhs(i,j),lhs(i,j)/rhs(i,j)
           endif
        enddo
     enddo
     stop

     do j=1,np
        do i=1,np
           if ( abs(lhs2(i,j)-rhs2(i,j)) .gt. 1d-20) then
              write(*,'(a)') 'ERROR: curl/curl_wk COV (2) integration by parts failure!'
              write(*,'(a,2i2,a,3e12.4)') '(i,j)=',i,j,' lhs,rhs=',lhs2(i,j),rhs2(i,j),lhs2(i,j)-rhs2(i,j)
           endif
        enddo
     enddo
  enddo

  print *,'done. integration by parts identity check:'
  end subroutine
end module
